package XML::ComicsML;

# $Id: ComicsML.pm,v 1.7 2001/04/17 04:22:03 jmac Exp $

# XXX TODO:
# default-guessing warning suppression option

use warnings;
use strict;
use utf8;

our $VERSION = '0.02';
our $AUTOLOAD;

# This first package is an abstract one, for all ComicsML objects.
use fields qw(language parent);

sub new {
    my $invocant = shift;
    my $class = ref($invocant) || $invocant;
#    my $self = {};
#    bless ($self,$class);
    my $self = fields::new($invocant);
    $self->_initialize(@_);
    return $self;
}

sub _initialize {
  my $self = shift;
  return $self;
}

my $minimum_version = '0.2';

######## Inheritable content-management stuff

sub AUTOLOAD {
  # Cheezo accessor
  my $self = shift;
  my ($attribute) = $AUTOLOAD =~ /^.*::(.*)$/;
  $attribute =~ s/-/_/g;
  if (@_) {
    $self->{$attribute} = $_[0];
  }
  return $self->{$attribute};
}

sub DESTROY {
  # Eliminate circular ref risks
  my $self = shift;
  delete($self->{parent});
}

# Langauge attributes are special, cuz they're scoped...
sub language {
  my $self = shift;
  if (@_) {
    $self->{language} = $_[0];
  }
  if ($self->{language}) {
    return $self->{language};
  } elsif ($self->parent) {
    return $self->parent->language;
  } else {
    return undef;
  }
}

sub remove_child {
  # Abstract way of taking an object and string identifying what sort
  # of object it is, and splicing it off an internal list of child
  # objects.
  my $self = shift;
  my ($type, $remove_me) = @_;
  my $children = $self->{$type}; # ref to internal list of this type of kids
  foreach (0..@$children) {
    if ("$$children[$_]" eq "$remove_me") {
     # found it... delete and return the removed object
      return (splice(@$children, $_, 1));
    }
  }
  return;                       # not found
}


########### Parser -- not methods

package XML::ComicsML::Parser;

use base qw(XML::Parser);

sub new {
    my $class = shift;
    my $self = $class->SUPER::new(
                                  Pkg=>'XML::ComicsML',
                                  ProtocolEncoding=>'UTF-8',
                                  Namespaces=>0,
                                  ErrorContext=>1,
                                  Handlers=>{
                                             Start=>\&start_handler,
                                             End=>\&end_handler,
                                             Char=>\&char_handler,
                                             Final=>\&final_handler,
                                            },
                                 );


    bless ($self,$class);
#    $self->_initialize(@_);
    return $self;
}

sub final_handler {
  my ($expat) = @_;
  return $expat->{current_comic};
}
  
sub start_handler {
  my ($expat, $element, %attrs) = @_;
  if ($element eq 'comic') {
    $expat->{current_comic} = XML::ComicsML::Comic->new;
    $expat->setHandlers(End=>\&XML::ComicsML::Comic::end_handler,
                        Start=>\&XML::ComicsML::Comic::start_handler,
                       );
    if ($attrs{'xml:lang'}) {
      $expat->{current_comic}->language($attrs{'xml:lang'});
    } else {
      $expat->xpcroak("Comic element needs a language attribute!!\n");
    }
    if ($attrs{version}) {
      if ($attrs{version} < $minimum_version) {
        $expat->xpcroak("Sorry, XML::ComicsML.pm supports version $minimum_version or later.\n");
      }
      $expat->{current_comic}->version($attrs{version});
    } else {
      $expat->xpcroak("Comic element needs a version attribute!!\n");
    }
  } else {
    $expat->xpcroak("Error: can only have comic elements at toplevel.\n\n");
  }
}

sub char_handler {
  my ($expat, $string) = @_;
  if ($string =~ /\S/) {
    $expat->{comic_textbuffer} ||= [];
    push (@{$expat->{comic_textbuffer}}, $string);
  }
}


package XML::Parser::Expat;

# Sort-of-a hacky way to get an accessor.

sub get_comic_buffer {
  my $self = shift;
  my @lines = $self->comic_buffer;
  # Clear out the textbuffer
  delete($self->{comic_textbuffer});
  # 'Fill' this text: Zap extra whitespace, then space-separate lines
  # and return them
  foreach (@lines) {
    s/^\s*(.*)\s*$/$1/;
  }
  return join(' ',@lines);
}

sub comic_buffer {
  my $self = shift;
  if ($self->{comic_textbuffer}) {
    return @{$self->{comic_textbuffer}};
  } else {
    return ();
  }
}

package XML::ComicsML::Comic;

use base qw(XML::ComicsML);
use fields qw(strips sorted_strips people version title last_built icon description url);
use XML::Writer;

sub _initialize {
  my $self = shift;
  $self->{strips} = {};
  $self->{sorted_strips} = [];
  $self->{people} = [];
  return $self;
}

sub people {
  # Returns list of person objects
  my $self = shift;
  if ($self->{people}) {
    return @{$self->{people}};
  } else {
    return ();
  }
}

sub add_person {
  # Add object representing <person> tag to this object.
  my $self = shift;
  my ($person) = @_;
  $self->{people} ||= [];
  push (@{$self->{people}}, $person);
}

sub new_person {
  my $self = shift;
  my $person = XML::ComicsML::Person->new;
  $person->parent($self);
  $self->add_person($person);
  return $person;
}

sub remove_person {
  my $self = shift;
  return $self->remove_child('people', @_);
}

sub set_defaults {
  # Guess and set values for required fields.
  # Also reports errors on stuff that's missing and which it can't guess.
  my $self = shift;
  unless ($self->language) {
    if (my $lang = $ENV{LANG}) {
      $lang =~ s/_.*$//;        # Strip region info
      $self->language($lang);
    } else {
      # This is weak...
      warn "WARNING: No language set on comic object, or in environment;\nassuming English.\n";
      $self->language('en');
    }
  }
  unless ($self->version) {
    warn "WARNING: No version set on comic object. Assuming $minimum_version.\n";
    $self->version($minimum_version);
  }
  unless ($self->last_built) {
    my @time = localtime;
    $time[5] += 1900;
    $time[4] += 1;
    my $stamp ="$time[5]-$time[4]-$time[3] $time[2]:$time[1]:$time[0]";
    warn "WARNING: No last-built time set on object; assuming $stamp.\n";
    $self->last_built($stamp);
  }
  unless ($self->title) {
    # XXX This sort of thing should carp, not die.
    die "ERROR: Comic object has no title set.";
  }
  if (my @people = $self->people) {
    foreach (@people) {
      $_->set_defaults;
    }
  } else {
    die "ERROR: Comic object requires at least one person defined.";
  }
  foreach ($self->strips) {
    $_->set_defaults;
  }
}

sub as_string {
  # Make this puppy printable.
  # Dumps on standard out for now...
  my $self = shift;
  $self->set_defaults;
  # @_ is a hash of XML::Writer args we'll pass right in.
  my $w = XML::Writer->new(DATA_MODE=>1, DATA_INDENT=>2, @_);
  $w->xmlDecl;
  $w->doctype('comic');
  $w->startTag('comic', version=>$self->version, 'xml:lang'=>$self->language);
  $w->dataElement(title=>$self->title);
  $w->dataElement('last-built'=>$self->last_built);
  $w->dataElement(icon=>$self->icon) if $self->icon;
  $w->dataElement(description=>$self->description) if $self->description;
  $w->dataElement(url=>$self->url) if $self->url;
  for my $person ($self->people) {
    $w->startTag('person');
    foreach (qw(firstname surname email url job)) {
      $w->dataElement($_=>$person->$_) if $person->$_;
    }
    $w->endTag('person');
  }
  for my $strip ($self->strips) {
    if ($strip->language ne $self->language) {
      $w->startTag('strip', id=>$strip->id, 'xml:lang'=>$strip->language);
    } else {
      $w->startTag('strip', id=>$strip->id);
    }
    $w->dataElement(title=>$strip->title) if $strip->title;
    $w->dataElement(url=>$strip->url) if $strip->url;
    $w->dataElement(date=>$strip->date);
    for my $person ($strip->people) {
      $w->startTag('person');
      foreach (qw(firstname surname email url job)) {
        $w->dataElement($_=>$person->$_) if $person->$_;
      }
      $w->endTag('person');
    }
    # Check for a teaser
    if ($strip->teaser_panels or $strip->description) {
      $w->startTag('teaser');
      $w->dataElement(description=>$strip->description) if $strip->description;
      $strip->panel_writer($w, 'teaser_panels');
      $w->endTag('teaser');
    }
    $strip->panel_writer($w, 'panels');
    $w->endTag('strip');
  }
  $w->endTag('comic');
  $w->end;
}

sub add_strip {
  my $self = shift;
  my ($strip) = @_;
  push (@{$self->{sorted_strips}}, $strip);
  if ($strip->id) {
    $self->{strips}->{$strip->id} = $strip;
  } else {
    # Not a very nice error message, dude.
    die "Bah, strip without id.\n";
  }
  $strip->parent($self);
}

sub new_strip {
  my $self = shift;
  my $strip = XML::ComicsML::Strip->new;
  my $id = (@_);
  $strip->id($id) if $id;
  $self->add_strip($strip);
  return $strip;
}

sub strips {
  my $self = shift;
  return @{$self->{sorted_strips}};
}

sub get_strip {
  # fetch a strip by its id.
  my ($self, $id) = @_;
  return unless $id;
  return ${$self->{strips}}{$id};
}

sub remove_strip {
  my $self = shift;
  my ($strip) = @_;
  # Remove from various places
  delete(${$self->{strips}}{$strip->id});
  return $self->remove_child('sorted_strips', $strip);
}

######## Parsing -- not methods

sub start_handler {
  my ($expat, $element, %attrs) = @_;
  if ($element eq 'strip') {
    my $strip = XML::ComicsML::Strip->new;
    $strip->parent($expat->{current_comic});
    $expat->{current_strip} = $strip;
    $strip->id($attrs{id});
    $expat->setHandlers(Start=>\&XML::ComicsML::Strip::start_handler,
                        End=>\&XML::ComicsML::Strip::end_handler,);
    if ($attrs{'xml:lang'}) {
      $strip->language($attrs{'xml:lang'});
    }
  } elsif ($element eq 'person') {
    my $person = XML::ComicsML::Person->new;
    $person->parent($expat->{current_comic});
    # XXX These handler resets can be more sexy, querying object
    # values to get the classnames, I betcha
    $expat->{current_person} = $person;
    $expat->setHandlers(Start=>\&XML::ComicsML::Person::start_handler,
                        End=>\&XML::ComicsML::Person::end_handler,);
  }
}

sub end_handler {
  my ($expat, $element) = @_;
  if ($element eq 'comic') {
    $expat->setHandlers(Start=>\&XML::ComicsML::start_handler,
                        End=>\&XML::ComicsML::end_handler,);
  } else {
    # abusing AUTOLOAD
    $expat->{current_comic}->$element($expat->get_comic_buffer);
  }
}

package XML::ComicsML::Person;

use base qw(XML::ComicsML::Comic);
use fields qw(firstname surname email job);

# Very simple package for the contents of <person> elements.

sub start_handler {
  # Nothing special here.
}

sub set_defaults {
  my $self = shift;
  unless ($self->surname) {
    # XXX do some system-level guessing here, maybe
    die "ERROR: person must have at least a surname\n";
  }
}

sub end_handler {
  my ($expat, $element) = @_;
  if ($element eq 'person') {
    $expat->{current_comic}->add_person($expat->{current_person});
    delete($expat->{current_person});
    $expat->setHandlers(Start=>\&XML::ComicsML::Comic::start_handler,
                        End=>\&XML::ComicsML::Comic::end_handler,);
  } else {
    # abusing AUTOLOAD
    $expat->{current_person}->$element($expat->get_comic_buffer);
  }
}

package XML::ComicsML::Strip;

use base qw(XML::ComicsML::Comic);
use fields qw(date panels teaser_panels id);

sub _initialize {
  my $self = shift;
  $self->{panels} = [];
  $self->{teaser_panels} = [];
  return $self;
}

sub set_defaults {
  my $self = shift;
  unless ($self->date) {
    # Hmm. Maybe we shouldn't guess this.
    my @time = localtime;
    $time[5] += 1900;
    $time[4] += 1;
    my $stamp ="$time[5]-$time[4]-$time[3]";
    warn "WARNING: No date set on strip ".$self->id."; assuming $stamp.\n";
    $self->date($stamp);
  }
  foreach ($self->teaser_panels, $self->panels, $self->people) {
    $_->set_defaults if defined($_) and $_;
  }
}

sub add_panel {
  my $self = shift;
  my ($panel) = @_;
  push (@{$self->{panels}}, $panel);
  $panel->parent($self);
}

sub new_panel {
  my $self = shift;
  my $panel = XML::ComicsML::Panel->new;
  $panel->parent($self);
  $self->add_panel($panel);
  return $panel;
}

sub panels {
  my $self = shift;
  return @{$self->{panels}};
}

sub remove_panel {
  my $self = shift;
  $self->remove_child('panels', @_);
}

sub new_teaser_panel {
  my $self = shift;
  my $panel = XML::ComicsML::Panel->new;
  $panel->parent($self);
  $self->add_teaser_panel($panel);
  return $panel;
}

sub add_teaser_panel {
  my $self = shift;
  my ($panel) = @_;
  push (@{$self->{teaser_panels}}, $panel);
  $panel->parent($self);
}

sub teaser_panels {
  my $self = shift;
  return @{$self->{teaser_panels}};
}

sub remove_teaser_panel {
  my $self = shift;
  return $self->remove_child('teaser_panels', @_);
}

sub panel_writer {
  # Internal method, called by the main package's output method.
  # Receives an XML::Writer object and a methodname for fetching a
  # type of panel, and bangs these things together for a while.
  my $strip = shift;
  my ($w, $method) = @_;
  if ($strip->$method) {
    $w->startTag('panels');
    for my $panel ($strip->$method) {
      if ($panel->language ne $strip->language) {
        $w->startTag('panel', 'xml:lang'=>$strip->language);
      } else {
        $w->startTag('panel');
      }
      $w->dataElement(url=>$panel->url) if $panel->url;
      $w->dataElement(caption=>$panel->caption) if $panel->caption;
      # This could probably be cleaner...
      my %descs = $panel->panel_descs;
      while (my($lang, $desc) = each(%descs)) {
        next if $lang eq 'default';
        if ($lang ne $strip->language) {
          $w->startTag('panel-desc', 'xml:lang'=>$lang);
        } else {
          $w->startTag('panel-desc');
        }
        for my $line ($desc->lines) {
          my $type = $line->type;
          my %attribs = ();
          # Set target and style attribs, if they exist for this part
          foreach (qw(target style)) {
            $attribs{$_} = $line->$_ if $line->can($_) and $line->$_;
          }
          $w->startTag($type, %attribs);
          foreach ($line->characters) {
            $w->dataElement(character=>$_);
          }
          $w->startTag('text') if $type eq "speech" or $type eq "thought";
          # Shut off data-mode so we can use mixed content
          $w->setDataMode(0);
          foreach ($line->text) {
            if ($$_[1] eq 'normal') {
              $w->characters($$_[0]);
            } else {
              $w->dataElement($$_[1]=>$$_[0]);
            }
          }
          $w->setDataMode(1);
          $w->endTag('text') if $type eq "speech" or $type eq "thought";
          $w->endTag($type);
        }
        $w->endTag('panel-desc');
      }
      $w->endTag('panel');
    }
    $w->endTag('panels');
  }
}

########## Parsing -- not methods

sub start_handler {
  my ($expat, $element, %attrs) = @_;
  if ($element eq 'panel') {
    my $panel = XML::ComicsML::Panel->new;
    $panel->parent($expat->{current_strip});
    $expat->{current_panel} = $panel;
    $expat->setHandlers(Start=>\&XML::ComicsML::Panel::start_handler,
                        End=>\&XML::ComicsML::Panel::end_handler,);
    if ($attrs{'xml:lang'}) {
      $panel->language($attrs{'xml:lang'});
    }
  } elsif ($element eq 'person') {
    my $person = XML::ComicsML::Person->new;
    $person->parent($expat->{current_strip});
    # XXX These handler resets can be more sexy, querying object
    # values to get the classnames, I betcha
    $expat->{current_person} = $person;
    $expat->setHandlers(Start=>\&XML::ComicsML::Person::start_handler,
                        End=>\&XML::ComicsML::Person::end_handler,);
  }
}

sub end_handler {
  my ($expat, $element) = @_;
  if ($element eq 'strip') {
    $expat->{current_comic}->add_strip($expat->{current_strip});
    delete($expat->{current_strip});
    $expat->setHandlers(Start=>\&XML::ComicsML::Comic::start_handler,
                        End=>\&XML::ComicsML::Comic::end_handler,);
  } elsif ($element eq 'teaser') {
    # Do nuthin...
  } else {
    # Let's try abusing AUTOLOAD...
    $expat->{current_strip}->$element($expat->get_comic_buffer);
  }
}

package XML::ComicsML::Panel;

use base qw(XML::ComicsML::Comic);
use fields qw(caption panel_desc);

sub _initialize {
  my $self = shift;
  $self->{panel_desc} = {};
}

sub caption {
  my $self = shift;
  if (@_) {
    $self->{caption} = $_[0];
  }
  return $self->{caption};
}

sub get_panel_desc {
  my $self = shift;
  my $language = $_[0] || 'default';
#  return @{$self->{panel_desc}{$language}};
  return $self->{panel_desc}{$language};
}

sub add_panel_desc {
  my $self = shift;
  # Two arguments are 0) a listref of the desc and 1) optionally, the
  # langauge encoding of this panel desc (such as 'en'). If no
  # langauge is supplied, it puts it in both the currently set
  # lanaguge and 'default'.
  if (my $language = $_[1]) {
    $self->{panel_desc}{$language} = $_[0];
  } else {
    $self->{panel_desc}{$self->language} = $_[0] if $self->language;
    $self->{panel_desc}{default} = $_[0];
  }
}

sub set_defaults {
  # Panels actually don't have defaults at this time...
}

sub new_desc {
  my $self = shift;
  my ($lang) = @_;
  my $desc = XML::ComicsML::Panel::Desc->new;
  $desc->parent($self);
  $self->add_panel_desc($desc, $lang);
  return $desc;
}

sub panel_descs {
  # Returns a hash of all panel desc objects, keyed to language.
  # Should this return a hashref instead? eh...
  my $self = shift;
  if (wantarray) {
    return %{$self->{panel_desc}};
  } else {
    # Scalar context: get the deafult desc only.
    return $self->get_panel_desc;
  }
}

sub remove_panel_desc {
  my $self = shift;
  my ($desc) = @_;
  return delete(${$self->{panel_desc}}{$desc->language});
}

######### Parsing -- not methods

sub start_handler {
  my ($expat, $element, %attrs) = @_;
  if ($element eq 'panel-desc') {
    if ($expat->{current_desc}) {
      $expat->xpcroak("Gack, already a current_desc\n");
    } else {
      $expat->{current_desc} = XML::ComicsML::Panel::Desc->new;
    }
  } elsif ($element eq 'thought' || $element eq 'speech' || $element eq 'narration' || $element eq 'action') {
    if ($expat->{current_desc}) {
      if ($expat->{current_desc_piece}) {
        $expat->xpcroak("Gack there's already a desc piece\n");
      } else {
        $expat->{current_desc_piece} = XML::ComicsML::Panel::Desc::Line->new;
        while (my($key, $val) = each(%attrs)) {
          $expat->{current_desc_piece}{$key} = $val;
        }
        $expat->{current_desc_piece}{type} = $element;
      }
    } else {
      $expat->xpcroak("Can't add new $element outside of panel-desc\n");
    }
  }
  # XXX Put updated error checks here
}

sub end_handler {
  my ($expat, $element) = @_;
  if ($element eq 'text' or $element eq 'narration' or $element eq 'action') {
    $expat->{current_desc_piece}->add_text($expat->get_comic_buffer) if $expat->comic_buffer;
# XXX put updated error checks here
    # XXX eww, this is yucky
  } elsif ($element eq 'strong' or $element eq 'emphasis' or $element eq 'interro' or $element eq 'bang' or $element eq 'interrobang' or $element eq 'silence') {
    $expat->{current_desc_piece}->add_text($expat->get_comic_buffer, $element);
  } elsif ($element eq 'character') {
    if ($expat->{current_desc_piece}) {
      $expat->{current_desc_piece}->add_character($expat->get_comic_buffer);
    } else {
      $expat->xpcroak("Can't define a character outside of a panel-desc piece.\n");
    }
  } elsif ($element eq 'panel-desc') {
    $expat->{current_panel}->add_panel_desc($expat->{current_desc});
    delete($expat->{current_desc});
  } elsif ($element eq 'panel') {
    # Decide whether is this is a strip-panel or a teaser-panel.
    if ($expat->within_element('teaser')) {
      $expat->{current_strip}->add_teaser_panel($expat->{current_panel});
    } else {
      $expat->{current_strip}->add_panel($expat->{current_panel});
    }
    delete($expat->{current_panel});
    $expat->setHandlers(Start=>\&XML::ComicsML::Strip::start_handler,
                        End=>\&XML::ComicsML::Strip::end_handler,);
  }
  # narration and action elements count as both desc_piece and
  # desc_text, so we do a separate if statement here.
  if ($element eq 'narration' || $element eq 'thought' || $element eq 'speech' || $element eq 'action') {
    $expat->{current_desc}->add_line($expat->{current_desc_piece});
    delete($expat->{current_desc_piece});
  }
}

package XML::ComicsML::Panel::Desc;

# class for panel-desc objects
use base qw(XML::ComicsML::Panel);
use fields qw(lines);

sub _initialize {
  my $self = shift;
  $self->{lines} = [];
  return $self;
}

sub new_line {
  my $self = shift;
  my $line = XML::ComicsML::Panel::Desc::Line->new;
  $line->parent($self);
  $self->add_line($line);
  return $line;
}

# Eh...

sub new_thought {
  my $self = shift;
  my $line = $self->new_line;
  $line->type('thought');
  return $line;
}

sub new_speech {
  my $self = shift;
  my $line = $self->new_line;
  $line->type('speech');
  return $line;
}

sub new_action {
  my $self = shift;
  my $line = $self->new_line;
  $line->type('action');
  return $line;
}

sub new_narration {
  my $self = shift;
  my $line = $self->new_line;
  $line->type('narration');
  return $line;
}

sub add_line {
  my ($self, $line) = @_;
  push (@{$self->{lines}}, $line);
}

sub lines {
  my $self = shift;
  return (@{$self->{lines}});
}

sub remove_line {
  my $self = shift;
  return $self->remove_child('lines', @_);
}

package XML::ComicsML::Panel::Desc::Line;

# abstract class for panel description lines

use base qw(XML::ComicsML::Panel::Desc);
use fields qw(type text characters target tone);

sub _initialize {
  my $self = shift;
  $self->{text} = [];
  $self->{characters} = [];
}

sub text {
  my $self = shift;
  return @{$self->{text}};
}

sub add_text {
  my $self = shift;
  my ($text, $type) = @_;
  $type ||= 'normal';
  push (@{$self->{text}}, [$text, $type]);
}

sub remove_text {
  my $self = shift;
  $self->remove_child('text', @_);
}

sub plaintext {
  # Convenience method; return the line's text as a scalar, ignoring
  # inline markup
  my $self = shift;
  my @text = $self->text;
  my @plaintext;                # return value
  foreach (@text) {
    # Do some conversion of vocal effects into ascii first...
    # XXX (yuck, ugly)
    if ($$_[1] eq 'interrobang') {
      push @plaintext, '?!';
    } elsif ($$_[1] eq 'bang') {
      push @plaintext, '!';
    } elsif ($$_[1] eq 'interro') {
      push @plaintext, '?';
    } elsif ($$_[1] eq 'silence') {
      push @plaintext, '...';
    } else {
      push @plaintext, $$_[0];
    }
  }
  foreach (@plaintext) {
    s/^\s*(.*)\s*$/$1/; # eliminate join-unfriendly whitespace
  }
  return join(' ',@plaintext);
}

sub add_character {
  # Associate a character (in the literary sense) to this line; note a
  # line can hold multiple characters.
  my $self = shift;
  my ($character) = @_;         # just a string, a character's name
  push (@{$self->{characters}}, $character);
}

sub characters {
  # Return list of all characters stuck to this object.
  my $self = shift;
  return (@{$self->{characters}});
}

sub remove_character {
  my $self = shift;
  return $self->remove_child('characters', @_);
}

1;
=pod

=head1 NAME

XML::ComicsML - create, write, and parse ComicsML files

=head1 SYNOPSIS

 use XML::ComicsML;
 
 # parse an existing ComicsML file
 my $parser = XML::ComicsML::Parser->new;
 my $comic = $parser->parsefile('my_comic.xml');
 
 my $title = $comic->title;
 print "The title of this comic is $title\n";

 my @strips = $comic->strips;
 print "It has ".scalar(@strips)." strips associated with it.\n";

 # build a comic object out of thin air, instead of parsing
 my $comic2 = XML::ComicsML::Comic->new;
 $comic2->title('Perl Adventures');
 $comic2->url('http://www.jmac.org/perlcomix/index.html');

 # define a person for it
 my $me = $comic2->new_person;
 $me->firstname('Emma');
 $me->surname('Shrimphaven');
 $me->email('emmashr@jmac.org');

 # add a strip
 my $strip1 = $comic2->new_strip('PA-1'); # My arbitrary ID for this strip
 $strip1->url('http://www.jmac.org/perlcomix/pa-1.html');
 $strip1->description('Captain Perl learns about list context.');
 
 # I could add some panel description here, but I am too lazy.
 # Let's just print out what we have to STDOUT.
 $comic2->as_string;

=head1 DESCRIPTION

This module provides a programmer interface for creating, parsing, and
otherwise working with documents written according to the ComicsML
format for describing digital comics, much as XML::RSS does for the
RSS syndication format.

You should really read the prose descriptions available about ComicsML
before attempting to use this module, or your head will explode. See
the links at the bottoem of this document for websites that explain
ComicsML far, far better than this manpage does, and the WARRANTY file
in case you ignore this precaution and your head explodes anyway.

=head2 Creating the main Comic object

The main object class that this module defines is
XML::ComicsML::Comic, which represents an entire document, and holds
various fields and other objects representing the document's
content. There are two main ways to create this object:

=over

=item * From a parser

One object class this module gives you, XML::ComicsML::Parser,
subclasses XML::Parser, so you may use all of that module's object
methods with this one, including the methods that parse existing
files, filehandles or scalars. The return value of all parsing methods
is an object belonging to the XML::ComicsML::Comic class.

The parser performs several checks specific to ComicsML validity as it
traverses a document, and will croak if it finds any errors.

=item * Roll your own

Alternatively, you may simply construct your own XML::ComicsML::Comic
object, and use its methods to fill in all its information. More
complex contents, such as strips, will need separate construction and
then insertion into the Comic object through the approproate method
(the B<new_strip> method, in a Strip object's case).

=back

Of course, you can combine these strategies, parsing an existing file
to create a Comic object with some content in it and then
manipulating, adding to, and subtracting from that content however you
like through the methods available through that object and all its
content objects.

=head2 Using it

A word about how all these object types fit together, before we dive
into all their methods: XML::ComicsML is a very container-oriented
module, and tries to follow the ComicsML spec as close as it can. The
following run-down of relationships should look somehow less confusing
if you're already familiar with the spec.

A Comic object (in other words, of the XML::ComicsML::Comic class --
I'll use abbreviated class names henceforth, because I'm lazy) can
hold several Strip objects, as well as multiple Person objects.

A Strip can also have Persons attached, and it can also have Panel
objects, of two flavors: "normal" panels, the ones that actually
comprise the strip itself, and teaser panels. Panels of either flavor
are instances of XML::ComicsML::Panel, but you access them through
different methods.

A Panel can have one Desc object for each lanaguge that panel has a
written description for.

Desc objects contain several Line objects, which in turn contain the
actual text of the description. Whew!

Finally, all of these objects have their own data fields as
appropriate -- for example, Comic objects have a place to store the
comic's title and URL. Additionally, all these sorts of containment
are purely optional; a Strip obejct is perfectly happy holding no
Panels at all. Again, refer to the ComicsML spec for more exact rules
on who holds what, how often, and why. :)

=head1 METHODS

=head2 XML::ComicsML::Parser Methods

=over

=item new

I<Class method.> Creates a new parser object.

=item parse

=item parsefile

=item ...and so on

XML::ComicsML objects inherit from XML::Parser, so all the lovely
XML-parsing methods those objects give you also exist here. See
L<XML::Parser> for more information on using them.

=back

=head2 XML::ComicsML::Comic Methods

From this point on, we'll use a lot of terms specific to ComicsML,
such as the particular things 'comic' and 'strip' mean in its context,
so we'll assume you're already familiar with them. If not, please
refer to the general ComicsML documentation, a URL to which you can
find in L<the SEE ALSO section of this manpage|"SEE ALSO">.

=over

=item new

I<Class method.> Creates a new Comic object.

=item new_strip

Creates a new XML::ComicsML::Strip object, adds it to the comic's list
of panels, and returns it to you. Under most circumstances, this
should be your Strip object constructor of choice.

=item remove_strip

When passed a Strip object as an argument, removes it from this
object's list of strips, and returns it if successful.

=item strips

Returns a list of all Strip objects attached to this Comic.

=item get_strip

Given a single argument representing the ID attribute of a strip
attached to this comic, returns that strip object.

=item new_person

Creates an XML::ComicsML::Person object as a child of this Comic, and
returns it to you.

=item people

Returns a list of all Person object attached to this Comic.

=item remove_person

When passed a Person object as an argument, removes it from this
object's list of descs, and returns it if successful.

=item version

=item language

=item title

=item last_built

=item icon

=item description

=item url

These methods are simple accessors: they each return the current value
of some field on the object, and if called with a single argument, set
it to that value first. Each relates to an attribute or sub-element of
the <comic> element within a ComicsML document. See the
documentation for ComicsML itself for more details about what sorts of
content they should contain.

I<Note about language>: An object's B<language> attribute corresponds
to the special "xml:lang" XML attribute, and ComicsML treats this as a
lexically scoped value. Therefore, if you set a value through the
language accessor of any object in the XML::ComicsML hiearchy, then
all the objects it contains will be able to read the same langauge
value through the B<language> accessor -- unless, of course, one of
these objects sets its own langauge value, which will in turn get
passed on to its own contents.

None of the other simple accessors anywhere else in the ComicsML
object tree are this wacky; they get or set values only as they might
appear on that one object, with no mind given to parents or
children. So there you have it.

=item as_string

Turns the Comic object into XML, courtesy of the XML::Writer
module. Uh, I did mention you need XML::Writer, right?

Passes along any arguments it recives as a hash to the underlying
constructor call to XML::Writer, so you can feed this method, for
example, a filehandle reference keyed under 'OUTPUT', and have the
ComicsML document wind up there.

By default, it prints to STDOUT, and uses a two-space indentation
style.

=back

=head2 XML::ComicsML::Strip Methods

=over

=item new_panel

Creates a new XML::ComicsML::Panel object, adds it to the end of the
comic's list of panels, and returns it to you. Under most
circumstances, this should be your Panel object constructor of choice.

The order in which a Strip object keeps its Panel objects is the same
as the order in which a human reader should encounter them -- in other
words, the order in which they would appear in an actual ComicsML
document.

=item new_teaser_panel

Just like B<new_panel>, except that the panel it returns will appear
inside the strip's teaser section, instead of its panels proper.

=item panels

Returns a list of all panel objects that live inside the main panels
section of this strip, in order of appearance.

=item remove_panel

When passed a Panel object as an argument, removes it from this
object's list of normal panels, and returns it if successful.

=item teaser_panels

Returns a list of all panel objects that live inside the teaser panels
section of this strip, in order of appearance.

=item remove_teaser_panel

When passed a Panel object as an argument, removes it from this
object's list of teaser panels, and returns it if successful.

=item new_person

Creates an XML::ComicsML::Person object as a child of this Strip, and
returns it to you.

=item people

Returns a list of all Person objects attached to this Strip.

=item remove_person

When passed a Person object as an argument, removes it from this
object's list of people, and returns it if successful.

=item title

=item url

=item date

=item description

Simple accessors that return the current value of the proper flavor of
strip attribute. Call with an argument to set it as such.

I<Note:> B<description> sets the description that will appear in the
strip's teaser section.

=back

=head2 XML::ComicsML::Panel Methods

=over

=item new_desc

Creates a new XML::ComicsML::Panel::Desc object, adds it to the comic's list
of panels, and returns it to you. Under most circumstances, this
should be your Panel::Desc object constructor of choice.

If called with an argument, it understands it as a language code, and
files this new panel-desc object under that language. If called with
no arguments, then the panel will file this desc under the same
language code as the one the panel had set as its own language.

=item panel_descs

If called in list context, returns a hash of all Panel::Desc objects
associated with this panel, keyed under their language codes. One of
them, the default desc, will show up twice, under the key
'default'. This is normal and beautiful.

If called in scalar context, returns only the default desc.

=item get_panel_desc

If called with a string argument, understands it as a language code,
and returns the Panel::Desc object filed in this panel under that
language.

If called with no arguments, returns the default Panel::Desc object.

=item remove_desc

When passed a Desc object as an argument, removes it from this
object's list of descs, and returns it if successful.

=item caption

=item url

Simple accessors that return the current value of the proper flavor of
strip attribute. Call with an argument to set it as such.

=back

=head2 XML::ComicsML::Panel::Desc Methods

=over

=item new_line

Creates a new XML::ComicsML::Panel::Desc::Line object, adds it to the
panel-desc's list of lines, and returns it to you. Under most
circumstances, this should be your Panel::Desc::Line object
constructor of choice.

=item new_thought

=item new_speech

=item new_action

=item new_narration

Convenience methods that work just like B<new_line>, but also set the
new Panel::Desc::Line object's type for you.

=item remove_line

When passed a Line object as an argument, removes it from this
object's list of lines, and returns it if successful.

=back

=head2 XML::ComicsML::Panel::Desc::Line Methods

Objects of this class hold the actual text of panel descriptions.

=over

=item type

Sets the type of this line; either I<speech>, I<thought>, I<narration>
or I<action>.

=item add_text

Called with one argument, adds it as unstyled text to this line
-- it's "normal" text that won't receive any special markup as XML.

Called with two arguments, adds the first as text with the second
argument understood as a style to apply to it -- in effect, adding an
XML inline element to this line.

For example:

 $line->add_text('Foo');

...will simply add 'Foo' as plain text to the current line. And:

 $line->add_text('Bar', 'emphasis');

...will "Bar" with an "emphasis" style to the current line. If you were to print this line as XML at this point, it would look like this:

 Foo<emphasis>Bar</emphasis>

=item text

Returns a list of anonymous two-element arrays holding all this line's
text and styles. If called on the line we've been building so far,
we'd get this:

 ( ['Foo', 'normal'],
   ['Bar', 'emphasis'] )

Note how unstyled text actually gets 'normal' as a default style.

=item plaintext

Returns a single string holding all text in this line, with its style
removed. If called on our exmaple line object we'd get:

 Foo Bar

Note that it automatically inserts a space in between text segments
(unless they had space already between them).

=item add_character

Only applicable with Desc objects of 'speech' or 'thought' type. Takes
a string argument, and adds a character with that name to the list of
characters speaking or thinking the current line.

=item characters

Returns a list of all characters (themselves represented by simple
scalars) speaking or thinking the current line.

=item target

=item tone

For 'speech'-type lines, fetches the current value of this attribute.
attribute. If called with a string argument, sets it to that first.

=back

=head2 XML::ComicsML::Person Methods

=over

=item firstname

=item surname

=item email

=item url

=item job

Simple accessors that return the current value of the proper flavor of
strip attribute. Call with an argument to set it as such.


=back


=head1 BUGS

Probably lots. Don't hesitate to email me if you find any, or see
sorely lacking features.

=head1 TODO

=over

=item *

Let the various sub-objects all have their own as_string methods.

=back

=head1 SEE ALSO

The ComicsML resource page: http://www.jmac.org/projects/comics_ml/

L<XML::Parser>

L<XML::Writer>

=head1 AUTHOR

Jason McIntosh <jmac@jmac.org>

=head1 COPYRIGHT

Copyright (c) 2001 by Jason McIntosh <jmac@jmac.org>.

This is free software; you can redistribute it and/or modify it under
the same terms as Perl itself.

=cut

